/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */
#define _GNU_SOURCE
#include <stdio.h>

#include <stdlib.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include "atsend.h"
#include "cmds.h"
#include "common.h"
#include "parse.h"

extern void closePort(int final);

extern int  	    serverMode;
extern char 	    callerId[MAXLEN];
extern int  	    repeatNow;
extern type_key*    repeatCmdPtr;

static void 	  updateTimerLastData (timerCmd* timer, char *data, int size);
static timerCmd*  createTimer(cmdItem *ci);
static int        processTimerCmd(cmdItem* ci);

char tmptext[MAXCMDLEN];
int  remoteOn       = 1;

timerCmd* timers  = NULL;

static char tofile[MAXLEN];

void setToFile()
{
	strcpy(tofile, "");
        char *e = getenv("TMPDIR");
        if (e) {
        	strcat(tofile, e);
        } else {
        	strcat(tofile, "/tmp");
        }
        strcat(tofile, TOFILE);
        e = getenv("USER");
        if (e) {
        	strcat(tofile, ".");
        	strcat(tofile, e);
        }        
        return;
}


// This is not clean-up function. It just finished the connection.
static void finishAll()
{
        // finish events forwarding
        printf("Finish events forwarding\n");
        logger("INFO","Finish events forwarding");

        if (serverMode == 0) {
	
		sendCMER(CMER_OFF);

                //handle_command(answer); // Does it needed ?
                // and resend No back
                //sendSyncCommand("AT+CKPD=\"e\"\r", answer, sizeof(answer), NULL);    // Does it needed ?
        }
        closePort(1);

        return;
}

static int execInFork(char *cmd)
{
	int ret;
        int cpid = fork();
        if( cpid < 0) {
        	logger("ERROR", "Can not fork!");
                return -1;
        }

        if(cpid) {      // father
        	waitpid(cpid,&ret,0);
                if(ret!=-1 && WIFEXITED(ret)) {
                	ret = WEXITSTATUS(ret);
                }
        } else {        // child
        	logger("INFO", "Close port in child");
                closePort(0);
                 
                ret = system(cmd);
                if(ret!=-1 && WIFEXITED(ret)) {
                	ret = WEXITSTATUS(ret);
                }
                exit(ret);
        }

        return ret;
}

static char *readResultsFile(int * count) 
{
	FILE *fp;
	struct stat buf;
	
        if(stat(tofile, &buf) == -1) {
		return NULL;
	}
	
	long fLen = buf.st_size;
	sprintf(tmp, "readResultsFile >%ld<", fLen);
	logger("DEBUG",tmp);
	
        fp=fopen(tofile,"r");
	if (fp == NULL) {
		return NULL;
	}
    
    	char * fBuffer = NULL;
    	fBuffer = calloc(fLen+1,1);
        if (fBuffer == NULL) {
        	logger("ERROR","no more memory!");
		return NULL;
        }
    
	(*count) = fread(fBuffer, sizeof(char), fLen, fp);
    	fclose(fp);
    
	//logger("DEBUG","readResultsFile: exiting");
	return fBuffer;
}

static char* substParams(char* in, cmdParams *params) {


        if (in == NULL) {
                logger("ERROR","substParams(): null input");
                return NULL;
        }
        sprintf(tmp, "substParams >%s<", in);
        logger("INFO",tmp);

        if (*in == '\0') { // Empty command. Skip it.
            return NULL;
        }
        
        char *command =  strdup(in);
        char *ptr     = command;
        char *papameterized = NULL;

        int pc = strlen(PARAM_CALLID);
        int pt = strlen(PARAM_TIME);
        int pm = strlen(PARAM_MODE);
        int pp = strlen(PARAM_PARAM);
        int pi = strlen(PARAM_INDEX);
        int pd = strlen(PARAM_CFGDIR);
        
        // calculate size to allocate
        int needSize = strlen(ptr)+1; 
        int nParams  = 0;
        while(ptr != NULL) {
                papameterized = strstr(ptr, PARAM_START);   // $(
                if (papameterized == NULL) { // Just copy
                        break;
                } else if(strncmp(papameterized+2,PARAM_CALLID,pc) == 0) {
                	needSize += strlen(callerId) - pc - 2;
                        ptr = papameterized + pc + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_TIME,pt) == 0) { 
               		needSize += 16 - pt - 2;
                        ptr = papameterized + pt + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_MODE,pm) == 0) {
                	char *m = getMode();
                        needSize += strlen(m) - pm - 2;	
                        ptr = papameterized + pm + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_PARAM,pp) == 0) {
                        if (params != NULL && params->value != NULL) {   // value can be empty
                        	needSize += strlen(params->value) - pp - 2;
                	}
                        ptr = papameterized + pp + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_INDEX,pi) == 0) {
                        if (params != NULL && params->index[0] != '\0') {
                        	needSize += strlen(params->index) - pi - 2;
                        }
                        ptr = papameterized + pi + 2;
                        nParams++;
                } else if (strncmp(papameterized + 2,PARAM_CFGDIR,pd) == 0) {
			int c = 0;
                        char *d;
                        if ((d = getCfgDir()) != NULL) {
				c = strlen(d);
                                needSize += c - pd - 2;
                        }
                        ptr = papameterized + pd + 2;
                        nParams++;
                } else {
                	break;
        	}
        }
        
        if (nParams == 0) {		// non-parametrized command
        	return command; 
        }

        papameterized = NULL;
        ptr = command;
        
        sprintf(tmp, "Allocate >%d< bytes for parametrized command", needSize);
        logger("DEBUG",tmp);

	char* cmd  = (char*) calloc(needSize+1,1);
       
        while(command != NULL) {
        	sprintf(tmp, "cur cmd >%s<", command);
                logger("DEBUG",tmp);

                papameterized = strstr(command, PARAM_START);   // $(
        
                if (papameterized != NULL) {
                	sprintf(tmp, "papameterized >%s<", papameterized);
                        logger("DEBUG",tmp);
                }
        
                if (papameterized == NULL) { // Just copy
                	logger("DEBUG", "no more parameters");
                        strcat(cmd, command);
                        break;
                } else if(strncmp(papameterized+2,PARAM_CALLID,pc) == 0) {
                	
                        *papameterized = '\0';
                        
                        strcat(cmd, command);		// First part
                        strcat(cmd, callerId);		// Add params
                
                        command = papameterized + pc + 2;
                    
                } else if (strncmp(papameterized + 2,PARAM_TIME,pt) == 0) { 
                	
                        *papameterized = '\0';

                        strcat(cmd, command);		// First part
                        time_t logtime;			// Add time
                        time(&logtime);

                        struct tm curtime;
                        localtime_r(&logtime, &curtime);

                        char buf[16];
                        sprintf(buf,"%d:%02d", curtime.tm_hour, curtime.tm_min);

                        strcat(cmd, buf);

                        command = papameterized + pt + 2;
                
                } else if (strncmp(papameterized + 2,PARAM_MODE,pm) == 0) {
                	
                        *papameterized = '\0';
                       
                        strcat(cmd, command); 		// First part
                        strcat(cmd, getMode());		// Add params
                                
                        command = papameterized + pm + 2;
                
                } else if (strncmp(papameterized + 2,PARAM_PARAM,pp) == 0) {
                	// This is dirty trick to make Bemused commands like VOLM<volume> works
                        // also it is used in Java Client list handling
                
                        *papameterized = '\0';

                        if (params == NULL) {   // value can be empty
                        	logger("WARNING", "No parameters supplied!");

                        	strcat(cmd, PARAM_START);
                        	strcat(cmd, PARAM_PARAM);
                        } else {
                        	strcat(cmd, command);		// First part
                        	strcat(cmd, params->value);	// Add params
                	}
                        command = papameterized + pp + 2;

                } else if (strncmp(papameterized + 2,PARAM_INDEX,pi) == 0) {
                	// Used in Java Client list handling
                
                        *papameterized = '\0';

                        if (params == NULL || params->index[0] == '\0') {
                        	logger("WARNING", "No parameters supplied!");

                        	strcat(cmd, PARAM_START);
                        	strcat(cmd, PARAM_PARAM);
                        } else {
                        	strcat(cmd, command);		// First part
                        	strcat(cmd, params->index);	// Add params
                        }
                        command = papameterized + pi + 2;

                } else if (strncmp(papameterized + 2,PARAM_CFGDIR,pd) == 0) {
                
                        *papameterized = '\0';
                
                        // First part
                        strcat(cmd, command);

                        // Add cfg. dir.
                        char *d;
                        if ((d = getCfgDir()) != NULL) {
				strcat(cmd, d);
                        }
                
                        command = papameterized + pd + 2;
                
                } else {    // Just copy
                	strcat(cmd, command);
                        break;
                }
        }
        int cl = strlen(cmd);
        sprintf(tmp, "final cmd >%s< %d", cmd,cl);
        logger("DEBUG",tmp);
        
        free(ptr);
        
        return cmd;
}

static int execSimpleCmd(char *cmd, int toFile)
{
	if (cmd == NULL) {
        	logger("DEBUG", "execSimpleCmd null input");
        	return EXIT_OK;
        }
        sprintf(tmp, "execSimpleCmd >%s<", cmd);
        logger("INFO",tmp);

        if (remoteOn != 1) {
                logger("DEBUG", "Remote if OFF. Skip event");
                return EXIT_OK;
        }

        int add = 0;
        if (toFile) { 
                add = strlen(tofile) + 4;
        } else {
        	add = 2;
        }
	char* p = calloc(strlen(cmd) + add,1);
        if (p == NULL) {
        	return EXIT_NOK;
        }
        strcpy(p,cmd);
         
        if (toFile) {       // We needs to wait results, so do not run in background
        	strcat(p, " > ");		
                strcat(p, tofile);
        } else {
        	if (! index(p, '&')) {
                	strcat(p, "&");
                }
        }
        int ret = execInFork(p);
        free(p);
        
        return (ret == -1 ? EXIT_NOK : EXIT_OK);
}


static int processSendCmd(cmdItem* ci)
{
	int ret = EXIT_OK;

        char all[MAXMAXLEN];
        strncpy(all, ci->descr, MAXMAXLEN-1);

        char *tag  = strtok(all, ",");
        if (tag == NULL) {
        	logger("DEBUG", "processSendCmd(): incorrect input (1)");                
        	return EXIT_NOK;
        }
        if (strlen(ci->descr) <= strlen(tag) + 1) {   // Set(tag,) or Set(tag) without any value
        	logger("DEBUG", "processSendCmd(): incorrect input (2)");                
                return EXIT_NOK;
        }
        char *ptr   = tag + strlen(tag) - 1;
        // strip spaces from tail
        while( *ptr == '\0' || *ptr == '\n' || *ptr == '\t' || *ptr == ' ') {
        	*ptr = '\0';
        	ptr--;
        }
        ptr = tag + strlen(tag) + 1;
        // strip spaces from head of value
        while( *ptr != '\0' && (*ptr == '\t' || *ptr == ' ')) {
        	ptr++;
        }

        sprintf(tmp, "processSendCmd() is >%s< >%s<",tag,ptr);
        logger("DEBUG", tmp);

        if (strcmp(tag, SEND_STRING) == 0) {
        	ret = writeSocketConn(ptr, strlen(ptr));
                if (ret != EXIT_OK) {
                	logger("DEBUG","Fails in writeSocketConn()");
                }
        } else if (strcmp(tag, SEND_BYTES) == 0) {
        	ret = writeBytesSocketConn(ptr);
                if (ret != EXIT_OK) {
                	logger("DEBUG","Fails in writeBytesSocketConn()");
                }    
        } else {
        	logger("DEBUG", "processSendCmd(): incorrect input (3)");                
                return EXIT_NOK;
        }
        return ret;
}

static void sendData(int type, int bytesFlag, char *data, int count)
{
        sprintf(tmp, "sendData() >%d<",count);
	logger("DEBUG",tmp);
        
	int ret = EXIT_OK;
	if (type == ID_SEND && bytesFlag) {
		ret = writeBytesSocketConn(data);
	} else if (count > 0) { // just check count to be sure we goes here from right place
		ret = writeSocketConn(data, count);
	}
	if (ret != EXIT_OK) {
		logger("DEBUG","Fails in write(Bytes)SocketConn()");
	}
}

static int macroCmd(char *macro, cmdParams* params) 
{
	logger("DEBUG","macroCmd");
        
        if (macro == NULL) {
        	return EXIT_NOK;
        }
 	int ret;
	type_key *k = findItem(macro, &ret, NULL);

	if (k && ret == FLAG_EXACT) {
		ret = handleCmdByKey(k, params);
		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
			return ret;
		}
	}
	return EXIT_OK;
}

void execDynamically(char *command) 
{
	cmdItem* tmpCmd = NULL;
	storeCmds(&tmpCmd, command); // parse command on-the-fly 
                
	processCommands(tmpCmd,NULL);
                
	freeCmds(tmpCmd);
	return;
}

static int loadCmds(char *file) 
{
	logger("DEBUG","loadCmds");
        if (file == NULL) {
        	return EXIT_NOK;
        }

	FILE *fp;
        char *command = NULL;
        size_t len = 0;
        ssize_t read;
	
	fp=fopen(file,"r");
	if (fp == NULL) {
		return EXIT_NOK;
	}
        
	while ((read = getline(&command, &len, fp)) != -1) {
        	sprintf(tmp,"Retrieved line %s of length %d ", command, (int) read);
		logger("DEBUG",tmp);
                
		execDynamically(command);

        }
        if (command) {
        	free(command);
        }
        fclose(fp);
	return EXIT_OK;
}

static int includeCmd(char *file) 
{
	logger("DEBUG","includeCmd");
        if (file == NULL) {
		return EXIT_NOK;
        }

        load_cfg(file, 0);
        
        //printKeys();
        printConf();
        
	sprintf(tmp,"Mode is %s", getMode());
	logger("DEBUG",tmp);
        
	return EXIT_OK;
}

static void uploadPix(char *cmdLine)
{
        sprintf(tmp, "uploadPix >%s<", cmdLine);
        logger("INFO",tmp);
	int prefixSz = 0;          		// image,icon,_name_, | _file_) or image,pix, | _file_)
        
        if (strstr(cmdLine,UPLOAD_PIX) != NULL) {
        	prefixSz = strlen(UPLOAD_PIX) + 1;
        } else {
        	int nCh = strlen(UPLOAD_ICON) + 1;
       		prefixSz = nCh + strstr(cmdLine+nCh,",") - (cmdLine+nCh) + 1;
	}
        
	char* fname = cmdLine + prefixSz;
        char* p = fname;
        while (*p != '\0') {	// remove trailing \n if exists
        	if (*p == '\n') {
                	*p = '\0';
                        break;
                }
                p++;
        }
        sprintf(tmp, "uploadPix file >%s<", fname); 
        logger("INFO",tmp);

	FILE *fp;
	struct stat buf;
	
        if(stat(fname, &buf) == -1) {
        	logger("ERROR","can't get file size!");
		return;
	}
	
	long fLen = buf.st_size;
	sprintf(tmp, "uploadPix >%ld<", fLen);
	logger("DEBUG",tmp);
	
        fp=fopen(fname,"r");
	if (fp == NULL) {
        	logger("ERROR","can't open file!");
		return;
	}
        
    	uint32_t szh32 = (uint32_t) fLen;
    	uint32_t szi32 = htonl(szh32);
    
    	char * fBuffer = NULL;
    	fBuffer = calloc(fLen+prefixSz+12,1);	// 11 =  "Set("     ,<size of bin data=4bytes>    ");\0"
        if (fBuffer == NULL) {
        	logger("ERROR","no more memory!");
		return;
        }
        
    	strcpy(fBuffer, "Set(");
    	strncat(fBuffer, cmdLine, prefixSz+1);
        memcpy((void*)fBuffer+prefixSz+4, (const void *) &szi32, 4);	// length on binary data
        
	fread(fBuffer+prefixSz+8, sizeof(char), fLen, fp);
    	fclose(fp);

        strcpy(fBuffer+prefixSz+fLen+8,");");
        
	int ret = writeSocketConn(fBuffer,prefixSz+fLen+10);
	if (ret != EXIT_OK) {
		logger("DEBUG","uploadPix(): Fails in writeSocketConn()");
	}

        free(fBuffer);
	
        return;
}

static int execCmdAndGetResults(int type, cmdItem* ci, timerCmd* timer, cmdParams *params) 
{
	logger("DEBUG","execCmdAndGetResults");

        if (serverMode == SERVER_NO || remoteOn != 1 || ci->exec == NULL) {
        	return EXIT_OK;
        }
        
        char* sExec = substParams(ci->exec, params);
        if (sExec == NULL) {
        	logger("DEBUG", "Empty exec. part after substParams()");
                return EXIT_NOK;
        }

        int ret = execSimpleCmd(sExec,1);
        if (ret != EXIT_OK) {
                sprintf(tmp, "Return code from execSimpleCmd() is %d",ret);
        	logger("DEBUG", tmp);
                free(sExec);
                return ret;
        }

        char* sDescr = substParams(ci->descr, params);
        if (sDescr == NULL) {
        	logger("DEBUG", "Empty command after substParams()");
                return EXIT_NOK;
        }
        
        int resLen = -1;
        char *res  = readResultsFile(&resLen);
	logger("DEBUG", "after readResultsFile()");

        if (timer != NULL) {
        	if (timer->lastValue != NULL && 
                    memcmp(timer->lastValue,res,timer->lastValueSize) == 0) {	
                	logger("DEBUG","No needs to send the same data.");
                	free(sDescr);
                	free(res);
                	return EXIT_OK;
                } else {
                	updateTimerLastData(timer, res, resLen);	// Update "last data"
                }
        }

        if (type == ID_SEND) {  // Just send string
        	
                int bytesFlag = 0;
                if (strstr(sDescr, SEND_BYTES)) {
                	bytesFlag = 1;
                }
                sendData(type, bytesFlag, res, resLen);
                
        } else {                // ID_SET
		logger("DEBUG", "ID_SET");

                if (strstr(sDescr,UPLOAD_ICON) != NULL || 
                    strstr(sDescr,UPLOAD_PIX)  != NULL) {
                                    
                    	char *chunk = calloc(strlen(sDescr)+resLen+2,1);
                        
                        strcpy(chunk,sDescr);
                        strcat(chunk,",");
                        memcpy(chunk+strlen(sDescr)+1,res,resLen);
                       
                	uploadPix(chunk);
                        
                        free (chunk);
                           
                } else {

                	// Replace all ");" inside buffer to "@@" since Java Client will separate commands by ");"
                	char *p;
                	while ((p = strstr(res,");")) != NULL) {	// binary data ???
                        	*p = '@';
                        	p++;
                        	*p = '@';
                	}
                
                	int memSize = resLen + strlen(sDescr) + 7;
                	char *chunk = calloc(1,memSize+1);
                	strcpy(chunk,CMD_SET);

                	strcat(chunk,"(");  // For example, Set(title
                	strcat(chunk,sDescr);
                
                	strcat(chunk,",");
                	memcpy(chunk+5+strlen(sDescr),res,resLen);	// could be binary data here
                	strcpy(chunk+5+strlen(sDescr)+resLen,");");
                	
                	//printf("CHUNK: %s\n", chunk);
                
                	sendData(type, 0, chunk, memSize);
                 
               		free (chunk);
        	}
  	}
        
	free(res);
	free(sDescr);

        logger("INFO","execCmdAndGetResults finished");
        return EXIT_OK;
}

static int processOneCommand(cmdItem* ci, cmdParams* p)
{
	 int ret;

         if (ci->exec == NULL || ci->descr == NULL) {
         	if (ci->descr == NULL) {
	         	sprintf(tmp, "processOneCommand>no description< >%d<", remoteOn);
                }        
         	if (ci->exec == NULL) {
	         	sprintf(tmp, "processOneCommand>no exec params< >%d<", remoteOn);
                }        
         } else {
         	sprintf(tmp, "processOneCommand >%s< >%s< >%d<", ci->descr, ci->exec, remoteOn);
         }
         logger("INFO",tmp);
         
         if (ci->type == ID_REMOTEON) {
                 logger("INFO", "Switch remote on");
                 remoteOn = 1;        
                 return EXIT_OK;
         }
         if (remoteOn != 1) {
         	logger("INFO", "Remote is off. Skip");
                return EXIT_OK;
         }
         
         switch (ci->type) {	// Non parametrized commands
         
         	case ID_EXIT:               
             		logger("INFO", "Exit key detected");
                    	finishAll();                        
			return EXIT_ABORT;
                 
         	case ID_TMPDISCONN:
                	logger("INFO", "Temporary disconnect key detected");
                 	finishAll();
                 	return EXIT_DISCON; // But what with the rest of the command line ? (now just skip)
                                  
             	case ID_REMOTEOFF:
               		logger("INFO", "Switch remote off");
               		remoteOn = 0;
                	return EXIT_OK;
                        
            	case ID_TIMER:   
                	logger("INFO", "Command: Timer");
                	if (serverMode != SERVER_NO) {
                		return processTimerCmd(ci);
                	}     
                	return EXIT_OK;

            	case ID_CANCELTIMER:   
                	logger("INFO", "Command: CancelTimer");
                	if (serverMode != SERVER_NO) {
                 		cancelTimer(ci->descr);
                	}  
                	return EXIT_OK;

            	case ID_SEND:
                	logger("INFO", "Command: Send");
                	if (serverMode != SERVER_NO) {
                        	ret = processSendCmd(ci);
                        	if (ret != EXIT_OK) {
                         		logger("DEBUG","Fails in processSendCmd()");
                        	}
                	}    
                	return EXIT_OK;
                        
                default:
                	break;        
        }
        
	switch (ci->type) {		// Parametrized commands, will parametrize in handler
                 
            	case ID_EXECSET: 
               		logger("INFO", "Command: ExecAndSet");
                	if (serverMode != SERVER_NO) { 
                		execCmdAndGetResults(ID_SET, ci, NULL, p);
               			logger("DEBUG", "Command: ExecAndSet FINISHED");
                        }
                	return EXIT_OK;
                 
            	case ID_EXECSEND:  
                	logger("INFO", "Command: ExecAndSend");
                	if (serverMode != SERVER_NO) {
               			execCmdAndGetResults(ID_SEND, ci, NULL, p);
                        }
                	return EXIT_OK;

            	case ID_MACRO:   
                	logger("INFO", "Command: Macro");
                 	ret = macroCmd(ci->descr,p);
               		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	return ret;
                        }
               		return EXIT_OK;
                 
                default:
                	break;        
        }

        if (ci->type == ID_EXEC) {	// Parametrized command, handler will use already parametrized data 
        	logger("INFO", "Command: Exec");
                
         	char* sExec = substParams(ci->exec, p);
        	if (sExec == NULL) {
        		logger("DEBUG", "Empty command after substParams()");
        		return EXIT_NOK;
        	}

                execSimpleCmd(sExec, 0);
                free(sExec);
                
                return EXIT_OK;
        }
        
        char* sDescr = substParams(ci->descr, p);
        if (sDescr == NULL) {
        	logger("DEBUG", "Empty command after substParams()");
                return EXIT_NOK;
        }
                       
	switch (ci->type) {		// Parametrized commands, handler will use already parametrized data 
                 
             	case ID_SETMODE:
                	logger("INFO", "Command: SetMode");
                         
                	int i = 0;    
                     	type_key* k = findItem(EVT_EXITMODE, &i, NULL);
                    	if (k && i == FLAG_EXACT) {
                                 logger("DEBUG", "Exec cmd on exit from mode");
                                 ret = handleCmdByKey(k,NULL);
                                 if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                                 	return ret;
                                 }
                      	} else {
                         	logger("DEBUG", "Leave mode");
                     	}
                      	setMode(sDescr);
                 
                     	i = 0; 
                     	k = findItem(EVT_ENTRMODE, &i, NULL);
                      	if (k && i == FLAG_EXACT) {
                                 logger("DEBUG", "Exec cmd on enter to mode");
                                 ret = handleCmdByKey(k,NULL);
                                 if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        		free(sDescr);
                                 	return ret;
                                 }
                   	} else {
                         	logger("DEBUG", "Entered to mode");
                       	}
                 	break;

             	case ID_SENDCKPD:
                	logger("INFO", "Command: SendCKPD");
                 	if (serverMode == SERVER_NO) {
                         
                        	ret = sendSeq(sDescr);
                        	if (ret != EXIT_OK) {
                         		logger("DEBUG","Fails in sendSeq()");
                         	}
                         	if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        		free(sDescr);
                         		return ret;
                         	}
                	}    
                 	break;
                 
            	case ID_SET:          
                	logger("INFO", "Command: Set");
                	if (serverMode != SERVER_NO && sDescr != NULL) {
                            
                        	strcpy(tmptext, CMD_SET);
                        	strcat(tmptext, "(");
                                
                 		//sprintf(tmp, "ID_SET >%s<", sDescr);
                 		//logger("DEBUG", tmp);
                                                 
                 		if (strstr(sDescr,UPLOAD_ICON) != NULL || 	// handle Set(image,icon|window,..._image_file_name_); separately.
                    		    strstr(sDescr,UPLOAD_PIX)  != NULL) {
                                    
                                    uploadPix(ci->descr);
                                    
                                } else {
                        		// Change \n in input to real '\n'
                        		char *p1 = sDescr;
                        		char *p2 = sDescr;
                        		while (p2 != NULL) {
                                		p2 = strstr(p1, "\\n");
                         
                                		if (p2 == NULL) {
                                 			strncat(tmptext, p1, p2-p1);
                                 		} else {
                                        		strncat(tmptext, p1, p2-p1);
                                        		strcat(tmptext, "\n");
                               			}
                                		p1 = p2 + 2;
                         		}
                        		strcat(tmptext, ");");
                 
                        		ret = writeSocketConn(tmptext, strlen(tmptext));
                        		if (ret != EXIT_OK) {
                         			logger("DEBUG","Fails in writeSocketConn() (Set)");
                         		}
                                }
            		}    
             		break;
                 
           	case ID_LOAD:   
               		logger("INFO", "Command: Load");
                 	ret = loadCmds(sDescr);
               		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	free(sDescr);
                        	return ret;
                        }
                 	break;

           	case ID_INCLUDE:   
               		logger("INFO", "Command: Include");
                 	ret = includeCmd(sDescr);
               		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	free(sDescr);
                        	return ret;
                        }
                 	break;
                 
		default:  
                	logger("INFO", "Unknown command");                 
                	break;
        }
	free(sDescr);
        
        return EXIT_OK;
}

int processCommands(cmdItem* cmdList,cmdParams* p) 
{
	while (cmdList) {
        	if (cmdList->descr == NULL) {
                        sprintf(tmp, "Process next command NULL parameters");
                } else {
                        sprintf(tmp, "Process next command %s",cmdList->descr);
                }
                logger("DEBUG", tmp);
                    
                int ret = processOneCommand(cmdList, p);
                if (ret != EXIT_OK) {
                	sprintf(tmp, "Return code from processOneCommand() is %d",ret);
                        logger("DEBUG", tmp);
                         
                        if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	logger("DEBUG", "Abort or disconnect after processOneCommand()");
                        
                                return ret;
                        }
                }
                cmdList = cmdList->next;
        }
        return EXIT_OK;
}

//
// Top-level command processing function
//
int handleCmdByKey(type_key* k, cmdParams* p)
{
	if (k == NULL) {
        	logger("DEBUG", "handleCmdByKey() NULL input");
                return EXIT_OK;
        }
        
        sprintf(tmp, "handleCmdByKey() >%s<",k->key);
        logger("DEBUG", tmp);
        
        if (p != NULL) {
	        sprintf(tmp,"handleCmdByKey params >%s,%s< ", p->index,p->value);
               	logger("DEBUG",tmp);
	}

        cmdItem* cmdList = getCommand(k);
        if (cmdList == NULL) {
        	logger("DEBUG", "No appropriate command was found by getCommand()");
        } else {
        	sprintf(tmp, "ready to execute command");
                logger("DEBUG", tmp);
        
                int ret = processCommands(cmdList, p);
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                	repeatNow    = 0;
                        repeatCmdPtr = NULL;
			logger("DEBUG", "handleCmdByKey() EXIT_ABORT || EXIT_DISCON");
                        return ret;
                }
        
                // In AT mode: User had pressed some key ... 
                // Send ToMainMenu sequence of CKPD's to show main menu again
                if (serverMode == SERVER_NO) {
                	if (sendSeq(getToMainMenu()) != EXIT_OK) {
                        	logger("DEBUG","Fails in sendSeq()");
                        }
                }
        }   

        return EXIT_OK;     
}

//
// Functions related to timer commands
//

static int processTimerCmd(cmdItem* ci)
{
	if (ci->exec == NULL || ci->tdata == NULL) {
        	logger("DEBUG", "processTimerCmd(): nothing to exec or timer data absent");
                return EXIT_NOK;
        }
        
        timerCmd* tc = NULL;
	if ((tc = createTimer(ci)) == NULL) {
        	logger("DEBUG", "processTimerCmd(): timer already exists or input was incorrect");                
                return EXIT_NOK;
        }

        execCmdAndGetResults(ID_SET, ci, tc, NULL);

        return EXIT_OK;
}

timerCmd* findTimer(char *id, timerCmd **prev) 
{
	if (id == NULL) {
        	return NULL;
        }

	timerCmd* ptr = timers;

        while (ptr != NULL && strncmp(ptr->timer->tdata->id,id,MTEXTLEN) != 0) {
        	if (prev != NULL) {
        		*prev = ptr;
                }
                ptr  = ptr->next;
        }

        return ptr;
}

static timerCmd* createTimer(cmdItem *ci)
{
	if (ci->tdata->id[0] == '\0' || ci->exec == NULL) {
        	logger("DEBUG", "createTimer(): wrong input");                
                return NULL;
        }

        if (findTimer(ci->tdata->id, NULL) != NULL) {
        	logger("DEBUG", "createTimer(): timer already exists");                
                return NULL;
        } 
        timerCmd *tm = calloc(1,sizeof(timerCmd));

        //tm->type      = type;
        tm->ticks     = 0;
        tm->times     = 1;
        tm->timer     = ci;
        tm->lastValue = NULL;
        tm->lastValueSize = -1;

        // Insert in head
        if (timers == NULL) { // Insert first
        	tm->next = NULL;
        } else {
        	tm->next = timers;
        }
        timers = tm;

        return tm;
}

int cancelTimer(char *id) 
{
	sprintf(tmp, "cancelTimer() >%s<", id);
        logger("DEBUG", tmp);

        timerCmd *prev = NULL;
        timerCmd *tm = findTimer(id, (timerCmd **) &prev);
        if (tm == NULL) {
            logger("DEBUG", "Cant find timer to cancel!");
            return EXIT_NOK;
        }    
        if (tm == timers) { //remove in head
        	timers = tm->next;
        } else {
        	if (prev != NULL) {
                	prev->next = tm->next;
                } else {
                	logger("ERROR", "Previous item absent for non-first element");
                        return EXIT_NOK;
                }
        }       
         
	tm->timer = NULL;
        free(tm->lastValue); tm->lastValue = NULL;
        free(tm);

        return EXIT_OK;
}

void freeTimers()
{
	logger("DEBUG", "freeTimers()");
        timerCmd *tm;

        while (timers != NULL) {
        	tm     = timers;
                timers = timers->next;
                
        	tm->timer = NULL;
                
                free(tm->lastValue); tm->lastValue = NULL;
                free(tm);
        }
        return;
}

int verifyTimers(int ticksInSec) 
{
	timerCmd *tm   = timers;

        while (tm != NULL) {
        	
                if (tm->ticks >= (tm->timer->tdata->timeout * ticksInSec)) {
                	sprintf(tmp, "verifyTimers(): it is time to execute >%s<", tm->timer->exec);
                        logger("DEBUG",tmp);

                        execCmdAndGetResults(ID_SET, tm->timer, tm, NULL);
                
                        tm->ticks = 0;
                
                        if (tm->timer->tdata->times > 0) {
                        	tm->times++;
                                if (tm->times >= tm->timer->tdata->times) {    // Cancel this timer 
                                	sprintf(tmp, "verifyTimers(): timer %s self-canceled", tm->timer->tdata->id);
                                        logger("DEBUG", tmp);
                                
                                        // Cancel this timer
                                        timerCmd *tmc = tm;
                                        tm = tm->next;

                                        cancelTimer(tmc->timer->tdata->id);
                                
                                        continue;
                                }
                        }
                } else {
                	tm->ticks++;
                }
                tm   = tm->next;
        }
	
        return EXIT_OK;
}

static void updateTimerLastData (timerCmd* timer, char *data, int size)
{
	if (timer->lastValue != NULL) {
        	free(timer->lastValue);
                timer->lastValue = NULL;
        }
        
        timer->lastValue     = data;
        timer->lastValue     = calloc(size + 1, 1);
        timer->lastValueSize = size;
        
        memcpy(timer->lastValue,data,size);
}
