/*
 * anyRemote
 * a bluetooth remote for your PC.
 * Copyright (C) 2006
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>


#include "common.h"
#include "parse.h"
#include "atsend.h"
#include "cmds.h"

#define TMPDISCONN_TIME 60
#define ALARM_CLEAN 	0
#define ALARM_CHECK 	1

static void mainRoutine     (void);
static int  isBemusedCommand(char *cmd);

char tmp	    [MAXMAXLEN];
char logfile	    [MAXLEN];
char callerId	    [MAXLEN];
char modifierString [MAXARGLEN];
char frontEndBuf    [MAXLEN];

int  alarmTimer   = 0;
int  repeatTimer  = 0;
int  callTimer	  = 0;
int  commandTimer = 0;
int  feTimer      = 0;

int  useCallId    = 1;
int  repeatNow    = 0;
type_key*  repeatCmdPtr = NULL;
int  callOn       = 0;

int  serverMode   = SERVER_NO;  // By default use AT commands and devices in /dev
                                // If set to 1 or 2, then do not use AT command and use socket 
int  port         = 0;
int  initialized  = 0;

void logger(char *head, char *str)
{
	FILE *fplog;
        time_t logtime;
        char *timestr;

	if (logfile[0] == '\0') {
        	return;
        }
        
        if (getLog()) {
        	
                if (strcmp(head, "DEBUG") == 0 && getDebug() == 0) {
                    return;
                }

                fplog = fopen(logfile, "a");
                if (fplog!=NULL) {
                	if (strcmp(head, "CFG") != 0) {
                        	time(&logtime);
                                timestr = (char *) ctime(&logtime);
                                timestr[strlen(timestr)-1]=0;
                                fprintf(fplog, "[%s] - ",timestr);
                        }   
                        if(head!=NULL) {
                        	fprintf(fplog, "%s - ",head);
                        }
                        fprintf(fplog, "%s\n",str);
                
                        fclose(fplog);
                } else {
                	printf("Can't open log file >%s<\n", logfile);
                }
        }
}

int openPort(char * portIn) 
{
	logger("INFO", "openPort()");
        char* strPort;
        int ret = 0;

        // device should be in format:
        // 1. /dev/something
        // 2. socket:5000
        // 3. bluetooth:10 or bluetooth (port will be 10 by default)

        if ((strPort = strstr(portIn, INET_SOCKET)) != NULL) {
        	
                strPort += strlen(INET_SOCKET);

                port = atoi(strPort);

                sprintf(tmp, "TCP Server mode. Use port %d\n", port);
                logger("DEBUG", tmp);

                if ((ret = openSocketPort(SERVER_TCP, port)) < 0) {
                	return -1;
                }

                serverMode = SERVER_TCP;

        } else if ((strPort = strstr(portIn, BT_SOCKET)) != NULL) {

        	strPort += strlen(BT_SOCKET);
        
                if (strstr(strPort, ":") == NULL) { // just "bluetooth
                	port = 10;
                } else {
                	strPort++;
                        port = atoi(strPort);
                }
                sprintf(tmp, "Bluetooth Sockets Server mode. Use channel %d", port);
                logger("DEBUG", tmp);

                if ((ret = openSocketPort(SERVER_BT, port)) < 0) {
                	return -1;
                }

                serverMode = SERVER_BT;

        } else {

        	sprintf(tmp, "Serial Client mode. Use device %s", portIn);
                logger("DEBUG", tmp);

                if (bt_open_port(portIn) < 0) {
                	printf("ERROR: open port\n");
                        return -1;
                }
        }

        if (initialized == 0) { 
        	// Run "onInit" command
                int i = 0;
                type_key *k = findItem(EVT_INIT, &i, NULL);
                if (k && i == FLAG_EXACT) {
                	logger("INFO", "Exec cmd on init");
                        handleCmdByKey(k,NULL);  // Since we not connected yet we did not interested in return code
                } else {
                	logger("INFO", "No command on init\n");
                }
                initialized++;
        }
        return 1;
}

void createDataDir() 
{
	char dd[542];
	char *t = getenv("HOME");
	strcpy(dd, t);
	strcat(dd, "/.anyRemote");

	DIR *d = opendir(dd);
	if (d) {
        	closedir(d);
        } else {
		mkdir(dd, S_IRWXU);
	}
}

// Then port will be closed from a forked child use 0
void closePort(int final) 
{
	if (serverMode > 0) {
         	closeSocketPort();
        } else {
        	if (bt_close_port(final) < 0) {
                	printf("ERROR: close port\n");
                }
        }

	if (final && getFrontEnd()) {
		sendToFrontEnd("Exiting");
		disconnectFrontEnd();
	}
}

int main(int argc, char *argv[])
{
	int i;
        for (i=1; i<argc; i++) {
        	if (strcmp(argv[i],"-f")==0 && i+1<argc) {
                	break;
                }
        }
	
        createDataDir();

        if (init_cfg(i<argc ? argv[++i] : NULL)!=1) {
        	printf("ERROR: Fails in init_cfg()\n");
                freeCfg();
                exit(1);
        }
                    
        if (parse_opts(argc, argv)!=1 ) {
        	printf("ERROR: Fails in parse_opts()\n");
                freeCfg();
                exit(1);
        }
        if (getLog()) {
        	// ALT
        	//strcpy(logfile, LOGFILE);
                char *t = getenv("TMPDIR");
                char *h = getenv("HOME");
                if (t) {
                	strcpy(logfile, t);
                } else {
                        if (h) {
                		strcpy(logfile, h);
                		strcat(logfile, "/.anyRemote");
                        } else {
                        	strcpy(logfile, "/tmp");
                        }
                }
                strcat(logfile, LOGFILE);
                if (t || (!h)) {	// add $USER suffix if not $HOME is used
                	t = getenv("USER");
                	if (t) {
                		strcat(logfile, ".");
                		strcat(logfile, t);
			}
                }
                printf("INFO: log file is %s\n",logfile);
                        
                // Just  truncate file
                FILE *fplog = fopen(logfile, "w");
                if (fplog) {
                	fclose(fplog);
                }      
                printConf();
        }

        // Not to core dump if connection will close
        signal(SIGPIPE, SIG_IGN);   
        signal(SIGCHLD, SIG_IGN);   

        // main cycle
        mainRoutine();

        return 0;
}

int handle_key_press(char* keyIn,int isRepeatable) 
{
	int ret;

        sprintf(tmp, "handle_key_press() >%s<", keyIn);
        logger("DEBUG", tmp);

        // Search an alias
        char *key = NULL;
        if((key = findAlias(keyIn)) == NULL) {
        	key = keyIn;
        }    
	
	if (getFrontEnd() == 1) {
		sendToFrontEnd(key);
	}

        // If modifier was set already it needs to verify which sequence we go
        if (modifierString[0] != '\0') {
        	strcat(modifierString," "); // Add one space first
                strcat(modifierString, key);
                sprintf(tmp, "Modifier+sequence is: >%s<", modifierString);
                logger("DEBUG", tmp);
        }

        // Execute appropriate command
        int i;
        cmdParams params;
        type_key *k = findItem((modifierString[0] == '\0' ? key : modifierString), &i, &params);
        if (k && i == FLAG_EXACT) {
        	if (isRepeatable && (!repeatNow) && getAutoRepeat()) {
                	logger("DEBUG", "Set auto repeat flag");
                        repeatNow    = 1;
                        repeatCmdPtr = k;
                } 
                ret = handleCmdByKey(k, NULL);
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) { 
			logger("DEBUG", "handle_key_press EXIT_ABORT || EXIT_DISCON");
                	return ret;
                }
                    
                // Clean modifier string
                modifierString[0] = '\0';
                
        } else if (k && i == FLAG_MULTIKEY){
        	logger("DEBUG", "Got part of multi keys sequence. Nothing to do.");
                if (modifierString[0] == '\0') {
                	logger("DEBUG", "Start of multi key sequence");
                        strcpy(modifierString, key);
                } 

                if(serverMode == SERVER_NO) {
                	ret = sendSeq(getToMainMenu());
                        if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                        	return ret;
                        }
                }
        } else if (k && i == FLAG_PARAMETR){
                logger("DEBUG", "Got parametrized command");
        
                ret = handleCmdByKey(k, &params);
                
                // Clean modifier string
                modifierString[0] = '\0';
		
                if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
                	 return ret;
                }
            
        } else {
        	// User had pressed some key ... 
                // Send ToMainMenu sequence of CKPD's to show main menu again
                if (serverMode == SERVER_NO && sendSeq(getToMainMenu()) != EXIT_OK) {
                	logger("DEBUG","Fails in sendSeq()");
                }

                logger("DEBUG", "No approprite key definition was found by findItem()");
                if (modifierString[0] != '\0') {
                	// Clean modifier string
                        modifierString[0] = '\0';
                        logger("DEBUG", "Clean modifier string");
                }
        }

        return EXIT_OK;
}

void cleanAutoRepeatFlag() 
{
	// Clear auto repeat flag
	logger("DEBUG","Clean auto repeat flag");
	repeatNow    = 0;
	repeatCmdPtr = NULL;
	repeatTimer  = 0;
}

int handle_command(char* cmd) 
{
	char *prev, *next, *key;
	int ret;

	if (cmd == NULL) {
		return EXIT_OK;
	}
	
	//skip lines starting with \n and \r
	if (cmd[0] == '\r') cmd++;
	if (cmd[0] == '\n') cmd++;

	sprintf(tmp, "handle_command() >%s<", cmd);
	logger("DEBUG", tmp);

	// most common case
	if (!cmd[0]) {
		//logger("DEBUG", "handle_command() skip 1");
		return EXIT_OK;
	}

	// if recieved multiline command - handle line by line and return
	prev = cmd;
	next = strchr(cmd, '\r');
	if (next == NULL && serverMode > 0) {	// Java client will send +CKEV: 1,1; +CKEV: 1,0
		next = strchr(cmd, ';');
	}
	if (next) {
		logger("DEBUG", "handle_command() multiline");

		char copy[1024];
		int len;

		do {	 
			len = next-prev;
	
			if (len > 2) {
				memcpy(copy, prev, len);
				copy[len] = 0;
			
				// use recursion
				ret = handle_command(copy);
				if (ret == EXIT_DISCON || ret == EXIT_ABORT) {
					return ret;
				}
			}
			prev = next+1;
			next = strchr(prev, '\r');
			if (next == NULL && serverMode > 0) {	// Java client will send +CKEV: 1,1; +CKEV: 1,0
				next = strchr(prev, ';');
			}
		} while (next) ;

		// now return
		return EXIT_OK;
	}

	if (IS_OK(cmd)) {   // OK - nothing to do
		return EXIT_OK;
	}


	if (memcmp(cmd, DEF_CLDEBUG, 12) == 0) {	            // Got debug event from Java client

		return EXIT_OK;
                
	} else if (memcmp(cmd, DEF_MSG, 4) == 0) {	    // Got event from Java client
		
		key = cmd+4;
		logger("DEBUG","Got event from client");
		return handle_key_press(key,0);

	} else if (memcmp(cmd, DEF_CKEV, 6) == 0) {
		    	
		// Handle different keypresses
		// Motorola: +CKEV: "1",1
		// SE	   : +CKEV: 1,1
		// Sagem: in rare ) cases it could be (between > and <) >,1:< ?

		int isSE = MODEL_SE;

		key = cmd+7;

		// Sagem test
		// key+=3;

		if (*key == '"') {	// Seems this is Motorola
			key++;
			//logger("DEBUG","+CKEV is in Motorola format");
			isSE = MODEL_MOTOROLA;
		}

		if (*key == ',') {	// Seems this is Sagem
			logger("DEBUG","+CKEV is in rare Sagem format (Empty code)");
			isSE = MODEL_SAGEM;
		}

		int i = 1;

		if (isSE == MODEL_SE) { 	// SonyEricsson & default
			while(key[i] != ',') i++;
			key[i] ='\0';
		
			i = i+1;
		} else if (isSE == MODEL_MOTOROLA) {	// Motorola
		    while(key[i] != '"') i++;
		    key[i] ='\0';

		    i = i+2;
		} else if (isSE == MODEL_SAGEM) {	// rare Sagem case
			*key ='\0';
		} else {
			logger("ERROR","Can't recognize +CKEV event !!!");
			return EXIT_NOK;
		}

		if (key[i] == '1') {
			logger("DEBUG","Button down event");
			return handle_key_press(key,1);
		} else if (key[i] == '0') {
			logger("DEBUG","Button up event");
			// In general we skip this
			if (repeatNow) {
				cleanAutoRepeatFlag();
			}
			return EXIT_OK;
		} else {
			sprintf(tmp, "something wrong: key should be pressed or released: cmd=>%s< char=>%c<",cmd, key[i]);
			logger("DEBUG",tmp);
			return EXIT_NOK;
		}
	} else if (memcmp(cmd, DEF_CLCC,6) == 0) {
	
		logger("INFO","Caller ID received. Skip it.");
		return EXIT_OK;
		
	} else if (strstr(cmd, DEF_RING) != NULL) {	// Incoming call
		logger("INFO","Incoming call"); 	    // This event sent periodically until used answered a call

		if (callOn == 0) {
			if (useCallId) {
				logger("DEBUG", "Get CallerId on incoming call");
				ret = getClip(callerId);

				sprintf(tmp, "CallerId is %s",callerId);
				logger("INFO", tmp);
		
				if(ret == EXIT_EXACT) { // Active call exists
					logger("DEBUG","Caller ID received");
				} else { 
					logger("DEBUG","Can't get Caller ID. Probably call was finished");
				}
			}
			//
			// May be we can get +CLIP: "+xxxxxxxxxxxx",145,,,,0< here ?
			// 

			int i = 0;
			type_key* k = findItem(EVT_INCALL, &i, NULL);

			if (k && i == FLAG_EXACT) {
				logger("DEBUG", "Exec cmd on incoming call");
				ret = handleCmdByKey(k,NULL);
				if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
					return ret;
				}
			}

			// If there is (EndCall) command,then set timer to determine when call ends
			i = 0;
			k = findItem(EVT_ENDCALL, &i, NULL);
			if (k && i == FLAG_EXACT) {
				logger("DEBUG", "Set call timer");
				callTimer = 0;
				callOn    = 1;
			}
		
			return EXIT_OK;
		} else {
			logger("DEBUG","Skip it."); // This event sent periodically until used answered a call
		}
	} else if (isBemusedCommand(cmd)) {	    // Bemused server emulation tricks
		return EXIT_OK;
	}

	sprintf(tmp, "Unhandled cmd %s", cmd);
	logger("DEBUG",tmp);

	return EXIT_OK;
}

static int read_command(char* buf, int l) 
{
	int ch, len = 0;
        char* p = buf;
        /*if (serverMode) {
        	len = readSocketConn(buf, l);
                if (len == EOF) {
                	buf[0] = 0;
                        return EOF;
                }
        } else {
        */ 
        	while (len < l) {
                	char c;
                	ch = bt_readchar(&c,100);
                        if (ch == EOF) {
                        	buf[0] = 0;
                                return EOF;
                        } else if (ch == EOF-1) {
                        	break;
                        } else if (ch >= 0 && (c == '\r' || c == ';')) {
                         	//sprintf(tmp, "Got break char >%c<", c);
				//logger("DEBUG",tmp);
                       		break;
                        } else  if (ch >= 0 && c != '\r' && c != ';') {	
                        	//sprintf(tmp, "Got next char >%c<", c);
				//logger("DEBUG",tmp);

                        	*p++ = c;
                                len++;
                        }
                }
        //}

        buf[len] = 0;
        return len;
}

// mode: 0 - clear alarms
//       1 - if fired - process, then clear alarms
void manageAlarms(int mode)
{
	char    *pf;
        cmdItem *pg;
        int i;
        struct stat buf;
            
        i = 0;
        while (1) {
        	
                //sprintf(tmp, "Call getAlarmInfo %d", i);
                //logger("DEBUG",tmp);
                pf = getAlarmInfo(i, &pg);

                if (pf == NULL) {
                        //logger("DEBUG", "After getAlarmInfo ALL PROCESSED");
                        break;
                }
                //sprintf(tmp, "After getAlarmInfo %s", pf);
                //logger("DEBUG", tmp);
                i++;
        
                // If file exists
                int fs = stat(pf, &buf);
                if (fs == 0) {
                	if (mode == ALARM_CLEAN) {
                        	logger("DEBUG","Remove file");
                        } else {            
                        	sprintf(tmp, "Alarm %s fired", pf);
                                logger("INFO",tmp);
                        
                                processCommands(pg, NULL);
                        }
                        remove(pf);
                }
        }
        return;
}

int initPort()
{
	char answer[1024];
        char cmd[32];

        logger("DEBUG","initPort");

        if (serverMode) {	// Wait for incoming connection
                logger("INFO", "Server mode: Waiting connection");
                return listenAndAcceptSocketConn();
        } 

        answer[0] = '\0';
        sendSyncCommand("ATZ\r", answer, sizeof(answer), NULL);
        //sendSyncCommand("AT\r", answer, sizeof(answer), NULL);
        if(!IS_OK(answer)) {
        	printf("ERROR: AT -> %s\n", answer);
                return -1;
        }

        sendSyncCommand("ATE0\r", answer, sizeof(answer), NULL);
        if(!IS_OK(answer)) {
        	printf("ERROR: ATE0 -> %s\n", answer);
                return -1;
        }
        char* ptr = getCharset();
        if (ptr) {
        	sprintf(cmd,"%s\"%s\"\r", DEF_AT_CSCS, ptr); 
                sendSyncCommand(cmd, answer, sizeof(answer), NULL);
                if(!IS_OK(answer)) {
                	printf("ERROR: Can't set charset to %s\n", ptr);
                }     
        }

        sprintf(cmd,"%s\r", DEF_AT_CLIP); 
        sendSyncCommand(cmd, answer, sizeof(answer), NULL);
        if(!IS_OK(answer)) {
            printf("ERROR: Can't set CLIP ON\n");
        }

        // Determine model
        sprintf(cmd,"%s\r", DEF_AT_CGMI); 
        sendSyncCommand(cmd, answer, sizeof(answer), NULL);

        //Set model in conf. data
        setModel(answer);

	int ret = sendCMER(CMER_ON);
        if(ret != EXIT_OK) {
        	printf("ERROR: fails in set event reporting on");
                return -1;
        }

        // Siemens S55 needs additional AT+CMEC=2 to make AT+CKPD works
        // not sure about other Siemens phones
        if (getModel() == MODEL_SIEMENS) {
        	sprintf(cmd,"%s\r", DEF_AT_CMEC); 
                sendSyncCommand(cmd, answer, sizeof(answer), NULL);
                if(!IS_OK(answer)) {
                	printf("ERROR: ON AT+CMEC\n");
                }
        }

        // Will set global callerId [MAXLEN];
        ret = getClip(callerId);
        if(ret == EXIT_NOK) {       // Got ERROR on AT+CLCC; probably phone does not supports this command
        	printf("ERROR: fails in getClip\n");
                useCallId = 0;
        }
        return 1;
}

static void freeAndExit(void) 
{
	logger("DEBUG","mainRoutine");
	
	freeTimers();
	freeCfg();
	exit(0);
}

static void mainRoutine(void) 
{
	char answer[MAXCMDLEN];
	int ret = EXIT_OK;
	int retRead, i;
	type_key *k = NULL;

	logger("DEBUG","mainRoutine");

	setToFile();

	int dmn = autoConnect();
	int rs  = getRetrySecs();
	strcpy(callerId, "NO CALLER ID");
	
	if (getFrontEnd()) {
		connectFrontEnd(-1);
	}
	
	while (1){
	    
		// try to open port
		if(openPort(getDevice()) == 1 &&   // Open device
		   initPort()	     == 1 ) {  // Init modem: AT, ATE0, AT+CMER, in server mode waits for connection
		
			manageAlarms(ALARM_CLEAN);
			
			if (getFrontEnd()) {
				sendToFrontEnd("Connected");
			}
	    	
			i = 0;    
	    		k = findItem(EVT_CONNECT, &i, NULL);
			if (k && i == FLAG_EXACT) {
		    		logger("INFO", "Exec cmd on connect");
		    		ret = handleCmdByKey(k,NULL);
		    		if (ret == EXIT_ABORT || ret == EXIT_DISCON) {
					return;
		    		}
			} else {
		    		printf("Start event forwarding\n");
			}

			logger("INFO","Start event forwarding");
		
			while (1) {	 
		    		//logger("DEBUG","main loop");
		    
	    	    		retRead = read_command(answer, sizeof(answer));
		    		if (retRead == EOF) {
	    				logger("INFO", "Got disconnected");
	    			
					if (getFrontEnd()) {
						sendToFrontEnd("Disconnected");
					}

					i = 0;
	    				k = findItem(EVT_DISCONNECT, &i, NULL);
					if (k && i == FLAG_EXACT) {
			    			logger("INFO", "Exec cmd on disconnect");
			    			ret = handleCmdByKey(k,NULL);
	    		    			// No additional handling since we already disconnected
					} 

	    				if (!dmn && !serverMode) {
			   	 		logger("INFO", "Exit because of disconnect");
			    			freeAndExit();  // Exit here
	    				}
	    		    
	    				if (serverMode == 0) {  // In server mode we will reuse already opened socket
	    		    			logger("INFO", "Closing the port");
			    			closePort(1);
					}   
	    				ret = EXIT_NOK;
	    		
	    				break;
		    		} else if (retRead > 0) {
					sprintf(tmp, "-------------------- Command read -------------------- >%s<", answer);
					logger("DEBUG",tmp);
	    	    
					ret = handle_command(answer);
	    		
	    				if (ret == EXIT_OK) {
	    		    			continue;
	    				} else if (ret == EXIT_ABORT) {
			   			 freeAndExit();  // Exit here
	    				} else if (ret == EXIT_DISCON) {
	    		    			break;
	    				}
	    	    		}
		    
		    		//Is any alarm fired (timeout about 15 seconds) ?
	    	    		if (alarmTimer == 300) {
	    				manageAlarms(ALARM_CHECK);
	    				alarmTimer = 0;
	    	    		} else {
	    				alarmTimer++;
	    	    		}

		    		// Is call still active (timeout about 1 seconds) ?
		    		if (callOn) {
	    				if (callTimer >= 20) {
	    		   			 logger("DEBUG", "Get call list again");
	    		    
	    		    			// Will set global callerId     [MAXLEN];
	    		    			ret = getClip(callerId);
	    		    			if (ret == EXIT_EXACT) {
	    						sprintf(tmp, "Call still active (%s)", callerId);
	    						logger("DEBUG",tmp);
	    		    			} else {
	    						logger("DEBUG","Call was finished");
	    						callOn = 0;
	    						strcpy(callerId, "NO CALLER ID");

	    						int i = 0;
	    						type_key *k = findItem(EVT_ENDCALL, &i, NULL);
	    						if (k != NULL && i == FLAG_EXACT) {
	    			    				logger("INFO","EVT_ENDCALL found");
	    			    				ret = handleCmdByKey(k,NULL);
	    			    				if (ret == EXIT_ABORT) {
									freeAndExit();      // Exit here
	    			    				} else if (ret == EXIT_DISCON) {
	    								break;
	    			    				}
	    						}
	    		    			}

	    		   			callTimer = 0;
	    				} else {
	    		    			callTimer++;
	    				}
	    	    		}
	    	    
	    	    		// Should we auto-repeat command (timeout about 1/10 second) ?
	    	    		if (repeatNow) {
					int isOdd = ((int)repeatTimer/2)*2;
					
					if (repeatTimer > 100) {  // Check for infinite loop. Only 50 autorepeats per one press
						cleanAutoRepeatFlag();
	    				} else if (repeatTimer > 0 && isOdd == repeatTimer) {
	    					sprintf(tmp, "Auto repeat command ... (%d)", repeatTimer);
	    					logger("DEBUG",tmp);
	    		    			ret = handleCmdByKey(repeatCmdPtr,NULL);
	    		    
	    		    			if (ret == EXIT_ABORT) {
							freeAndExit();      // Exit here
	    		    			} else if (ret == EXIT_DISCON) {
	    						break;
	    		    			}
	    		    			//repeatTimer = 0;
						repeatTimer++;
	    				} else {
						//logger("DEBUG","Auto repeat command ...Skip");
	    		    			repeatTimer++;
	    				}
	    	    		}

		    		// Verify commands executed by timer (timeout about 1/2 sec)
	    	    		if (commandTimer == 10) {
	    				//logger("DEBUG","Verify timer commands ...");
	    				ret = verifyTimers(1) ;
	    				commandTimer = 0;
					
	    	    		} else {
	    				commandTimer++;
	    	    		}
				
		    		// Verify commands executed by timer (timeout about 1/2 sec)
				if (getFrontEnd() == 1) {
	    	    			if (feTimer == 10) {
	    					//logger("DEBUG","Verify front end timer...");
					
						ret = readFromFrontEnd(frontEndBuf, MAXLEN);
						if (ret == EXIT_OK) {
							sprintf(tmp, "readFromFrontEnd() >%s<", frontEndBuf);
	    						logger("DEBUG",tmp);
							
							execDynamically (frontEndBuf);
							frontEndBuf[0] = '\0';
							
						//} else {
						//	logger("DEBUG","readFromFrontEnd() EXIT_NOK");
						}	
	    	    			} else {
	    					feTimer++;
	    	    			}
				}
	    	    
		    		//if (getKeepAlive()) {
				//	sendKeepAlive();
				//}
				
	    	    		// Main loop timer (1/20 of second)
		    		usleep(50000);
		 	}  
		}

		// Can't open port or it closed again
		if (dmn || serverMode > 0 || ret == EXIT_DISCON) {
	    		int timeout;
	    	
	    		if (ret == EXIT_DISCON) {
	    	   		timeout = TMPDISCONN_TIME;
	    	    		ret = EXIT_OK;
	    		} else {
	    	    		timeout = rs;
	    		}

	    		sprintf(tmp, "Wait %d seconds to connect ...", timeout);
	    		logger("INFO",tmp);
	    		printf("INFO: Wait %d seconds to connect ...\n", timeout);
	    	
	    		sleep(timeout);
	    	
	    	} else {	// Proceed to exit
	    		break;
	    	}
	}

	// Finish all
	logger("INFO", "Stop process ...");
	closePort(1);
	
	return;
}

//
// Bemused server limited emulation
// we could get more than one command at once
//
      
static int isBemusedCommand(char *cmdIn)
{  
	sprintf(tmp, "isBemusedCommand: >%s<",cmdIn);
        logger("DEBUG",tmp);

        int isBemused = 0;

        char *cmd = cmdIn;
        char *last = cmdIn + strlen(cmdIn);

        while (cmd < last) {
        	int handle = 0;
                int shift  = 4;

                char oneCmd[10];
                char paramValue[8];
                paramValue[0] = '\0';

                if (strncmp(cmd, "CHCK", 4) == 0 ||
                    strncmp(cmd, "DINF", 4) == 0 ||
                    strncmp(cmd, "EXIT", 4) == 0 ||
                    strncmp(cmd, "FADE", 4) == 0 ||
                    strncmp(cmd, "FFWD", 4) == 0 ||
                    strncmp(cmd, "GVOL", 4) == 0 ||
                    strncmp(cmd, "INF2", 4) == 0 ||
                    strncmp(cmd, "INFO", 4) == 0 ||
                    strncmp(cmd, "LIST", 4) == 0 ||
                    strncmp(cmd, "NEXT", 4) == 0 ||
                    strncmp(cmd, "PAUS", 4) == 0 ||
                    strncmp(cmd, "PLEN", 4) == 0 ||
                    strncmp(cmd, "PLST", 4) == 0 ||
                    strncmp(cmd, "PREV", 4) == 0 ||
                    strncmp(cmd, "RMAL", 4) == 0 ||
                    strncmp(cmd, "RWND", 4) == 0 ||
                    strncmp(cmd, "SHUT", 4) == 0 ||
                    strncmp(cmd, "STEN", 4) == 0 ||
                    strncmp(cmd, "STOP", 4) == 0 ||
                    strncmp(cmd, "STRT", 4) == 0 ||
                    strncmp(cmd, "VERS", 4) == 0) {
                	handle = 1;
                }

                if (strncmp(cmd, "REPT", 4) == 0 ||
                    strncmp(cmd, "SHFL", 4) == 0) {
                
			// Just "toggle" functionality
                        //unsigned char bc = ((unsigned char) *(cmd + 4));
                        //sprintf(paramValue, "%d",bc);
                
                        handle = 1;
                        shift  = 5;
                }
                if (strncmp(cmd, "VOLM", 4) == 0) {
                	unsigned char bc;
                        if (strlen(cmd) == 4) {  // read only VOLM without value to set
                        	logger("DEBUG","read one more byte");
                                char ret;
                                //if ((ret = readSocketConn((char*)buf, 2)) != 1) {
                                char c;
                                if ((ret = bt_readchar(&c,500000)) < 0) {
                                	sprintf(tmp, "isBemusedCommand: read < 0");
                                        logger("DEBUG",tmp);
                                        return 0;
                                }
                                sprintf(tmp, "isBemusedCommand: read >%c<", c);
                                logger("DEBUG",tmp);
                                bc = ((unsigned char) c);
                        } else {
                        	bc = ((unsigned char) *(cmd + 4));
                        }

                        // VOLM 0-255 -> 0%-100%
                        sprintf(paramValue, "%d", ((int)bc * 100)/255);
                
        
                        sprintf(tmp, "isBemusedCommand: VOLM parameter >%s<", paramValue);
                        logger("DEBUG",tmp);
                
                        handle = 1;
                        shift  = 5;
                }
                if (strncmp(cmd, "SLCT", 4) == 0) {
                	handle = 1;
                        shift  = 6;
                }
                if (strncmp(cmd, "SEEK", 4) == 0)  {    

                	sprintf(paramValue, "%d", 
                                     (((unsigned char)*(cmd + 4)) << 24) + 
                                     (((unsigned char)*(cmd + 5)) << 16) + 
                                     (((unsigned char)*(cmd + 6)) << 8)  + 
                                       (unsigned char)*(cmd + 7));
                
                        handle = 1;
                        shift  = 8;
                }
                if (strncmp(cmd, "DLST", 4) == 0 || 
                    strncmp(cmd, "DOWN", 4) == 0 || 
                    strncmp(cmd, "FINF", 4) == 0 ||    
                    strncmp(cmd, "LADD", 4) == 0 ||
                    strncmp(cmd, "PLAY", 4) == 0) {             
                    
                	unsigned char bc = *((unsigned char*) (cmd + 4));
                        sprintf(paramValue, "%d",bc);
                
                        shift  = 5 + bc;
                        handle = 1;
                }
                if (handle) {
                	isBemused = 1;
                        strncpy(oneCmd,cmd,4); 
                        oneCmd[4] = '\0';
                        if (paramValue[0] != '\0') {
                        	strcat(oneCmd,"(-1,");
                                strcat(oneCmd,paramValue);
                                strcat(oneCmd,")");
                        }
                
                        sprintf(tmp, "isBemusedCommand: one commmand >%s<", oneCmd);
                        logger("DEBUG",tmp);

                
                        handle_key_press(oneCmd,1);
                } else {
                	break;
                }

                cmd = cmd + shift;
        }
        return isBemused;
}
